#!/usr/bin/tclsh
#
#  консольная программа
#  для дистрибуции скриптов-индикатров-советников-библиотек MQL4 
#  ---------------
#  изначально сделана для создания zip диструбутивов ATcl
#

#  из-за "особенностей" MetaTrader отдельные компоненты программ 
#  разносятся в разные каталоги относительно "каталога данных":
#  MQL4/Includes/${product}.mqh - основноый заголовочный файл 
#  MQL4/Includes/${product}/*.* - все прочие include
#
#  MQL4/Experts/${product}*.mq{4|h} - исходники основных экспертов/роботов и их include
#  MQL4/Experts/${product}*.ex4 - их "скомпилированные" бинарники
#  MQL4/Experts/${product}/* - все побочные/дополнительные/производные эксперты
#
#  и так далее с каталогами Experts,Includes,Libraries,Scripts 
#  плюс в Libraries могут быть ещё dll
#
#  цель программы - обеспечить пакетную сборку зависимостей и построить архив со всеми файлами программы
#

set productName ""
set dataDir ""
array set productFiles {}

proc mq45dist { {fileOrDir {}} } {

	global dataDir
	global productName

	DetermineProductAndDataDir $fileOrDir
	if { $productName == "" } {
		puts stderr "ProductName not determinated"
		return;
	}
	if { $dataDir == "" } {
		puts stderr "DataDir not determinated"
		return;
	}
	BuildFileList
	MassCompile
	ZipProduct
	
}

proc DetermineProductAndDataDir { {fileOrDir {}} } {

	global dataDir
	global productName
	
	set dataDir 	""	;	# каталог данных терминала
	set productName ""	;	# имя продукта
	if { $fileOrDir == {} || $fileOrDir == "." } {
		# если ничего не задано, начинаем поиск с текущего каталога
		set fileOrDir [ pwd ]
	}
	if { [ file isfile $fileOrDir ] } {
		# на входе - обычный файл, не каталог
		# определим абсолютное имя
		set owd [ pwd ]
		cd [ file dirname $fileOrDir ]
		set baseDir [ pwd ]
		cd $owd
		set startDir $baseDir
		# попробуем определить имя продукта из него
		set path [ file split $startDir ]
		if { [ string tolower [ lindex $path end ] ] in {Experts Scripts Includes Libraries} } {
			set productName [ file rootname [ lindex $path end ] ]
		}
	} elseif { [ file isdirectory $fileOrDir ] } {
		# нам нужно абсолютное полное имя
		set owd [ pwd ]
		cd $fileOrDir
		set startDir [ pwd ]
		cd $owd
	} else {
		# с прочими спец.файлами - не работаем
		puts "WTF?"
		return
	}
	set dirs [ file split $startDir ]
	# просматриваем каталоги от конца к корню
	set i [ llength $dirs ]
	for {incr i -1 } { $i > 0} { incr i -1 } {
		set name [ lindex $dirs $i ]
		set lowerName [ string trim [ string tolower $name ] ]
		if { $lowerName in {config history logs mql4 mql5 profiles sounds templates tester} } {
			# возможно достигли корня иерархии dataDir
			set path [ file join {*}[lrange $dirs 0 [ expr $i -1 ] ] ]
			if { [ file exists  [ file join $path config terminal.lic ] ] } {
				puts "productName: $productName" 
				puts "DataDir: $path" 
				set dataDir $path
				return {}
			}
			continue
		}
		if { $lowerName in {lib libs var etc src app tmp inc include} } {
			# общеупотребительные имена каталогов, не могут быть именем продукта
			continue
		}
		if { $lowerName in {experts files images includes libraries logs presets projects scripts market} } {
			# имя не может быть типичным для MQL4/* 
			continue
		}
		if { $lowerName in {caches files history logs} } {
			# имя не может быть типичным для tester/* 
			continue
		}
		if { $lowerName in {default deleted downloads mailbox signals symbolsets} } {
			# имя не может быть типичным для history/* 
			continue
		}
		### похоже найдено требуемое имя ###
		set productName $name
	}
}
proc BuildFileList {} {
	update
	puts "Build FileList"
	global dataDir
	global productName
	
	set owd [ pwd ]
	cd $dataDir

	global dataDir
	global productName
	global productFiles
	
	foreach dir { mql4 mql5 profiles sounds templates tester } {
		if { [ file isdir $dir ] } {
			FindProductFiles $dir
		}
	}
	cd $owd
}
proc PrintFileList {} {
	global productFiles
	foreach name [ lsort -dictionary [ array names productFiles ] ] {
		puts $name
	}
}
# рекурсивный обход каталогов и выявление файлов продукта
proc FindProductFiles { dir } {
	update
	global dataDir
	global productName
	puts "Find product files in $dir"
	foreach path [ glob -nocomplain -directory $dir * ] {
		set name [ file tail $path ]
		#puts "Check $name"
		if { [ file isfile $path ] } {
			if { [ string match ${productName}* $name ] } {
				AddFile $path
			}
		} elseif { [ file isdirectory $path ] } {
			if { [ string match ${productName}* $name ] } {
				AddDirectory $path
			} else {
				if { [ string tolower $name ] ni { market } } {
					FindProductFiles $path
				}
			}
		}
	}
}
# добавить файл к списку файлов продукта
proc AddFile { path } {
	update
	global productFiles 
	if { [ string tolower [ file extension $path ] ] in { .out .tmp .log } } {
		# не включать временные файлы
		return
	}
	set productFiles($path) [ file mtime $path ]
}
# добавить каталог к списку файлов
proc AddDirectory { dir } {
	update
	foreach path [ glob -nocomplain -directory $dir * ] {	
		set name [ file tail $dir ]
		if { [ file isfile $path ] } {
			AddFile $path
		} elseif { [ file isdirectory $path ] } {
			AddDirectory $path
		}
	}
}
proc MassCompile {} {
	global productFiles
	global dataDir
	set owd [ pwd ]
	cd $dataDir
	set latest 0
	set toCompile {}
	puts "MassCompile"
	foreach { name mtime } [ array get productFiles ] {
		set ext [ file extension $name ] 
		set lowerExt [ string tolower $ext ]
		if { $lowerExt ni { .ex4 .ex5 } } {
			if { $mtime > $latest } {
				set latest $mtime
			}
		} 
		if { [ string tolower $ext ] in { .mql .mqh .mq4 .mq5 } } {
			lappend toCompile $name
		}
	}
	foreach name $toCompile {
		file mtime $name $latest
		Compile $name
	}
	cd $owd
}

proc Compile { filename } {
	global dataDir
	global productFiles
	puts "Compile $filename"
	set compiler [ list [ file join $dataDir metaeditor ] /portable ]
	set cmdline {/log}
	set ext [ string tolower [ file extension $filename ] ]
	set bin {}
	if { $ext in { .mq4 } } {
		# ожидаем бинарник ex4
		set bin [ file rootname $filename ].ex4
		# удалить ex4 если он был
		file delete $bin
		catch { unset productFiles($bin) }
		# каталог include
		set include [ file join $dataDir mql4 includes ]
		# добавить аргументов 
		lappend cmdline "/include:$include" 
		set source [ file join $dataDir $filename ]
		set cmdline [ list {*}$compiler "/compile:$source" {*}$cmdline ]
	} elseif { $ext in { .mq5 } } {
		# ожидаем бинарник ex5
		set bin [ file rootname $filename ].ex5
		# удалить ex4 если он был
		file delete $bin
		# каталог include
		set include [ file join $dataDir mql5 includes ]
		# добавить аргументов 
		lappend cmdline "/compile:$filename"
	} elseif { $ext in { .mqh .mql } } {
		# бинарников нет
		set bin {}
		set source [ file join $dataDir $filename ]
		set cmdline [ list {*}$compiler "/compile:$source" /s {*}$cmdline ]
	}
	puts [ join $cmdline " " ]
	set out [ open |$cmdline "r" ] 
	while { ! [ eof $out ] } {
		gets $out
		update
	}
	catch { close $out }
#	if { [ catch { exec {*}$cmdline } msg ] } {
#		set out [ lrange [ split $msg "\n" ] 1 end ]
#		if { [ llength $out ] > 0 } {
#			puts "  "[ join $out "\n  "]
#		}
#	}
	if { $bin != {} } {
		# должен получиться бинарник
		if { [ file exists $bin ] } {
			puts "Add binary : $bin"
			set productFiles($bin) [ file mtime $bin ]
		} else {
			puts "!!! Compilation fault, $bin not found"
		}
	}
}
proc ZipProduct {} {
	global dataDir
	global productName
	global productFiles
	# получить макс.время (могло поменяться при компиляции)
	set latest 0
	foreach { name mtime } [ array get productFiles ] {
		if { $mtime > $latest } {
			set latest $mtime
		}
	}
	# перестраховка
	if { $latest == 0 } {
		set latest [ clock seconds ]
	}
	# сформировать имя архива
	set time [ clock format $latest -format "%Y%m%d" ]
	set zipName ${productName}-${time}.zip
	# список файлов через \n
	set fileList [ file join [ lsort -dictionary [ array names productFiles ] ] "\n" ]
	# командная строка
	set cmdline [ list zip -v -@ -9 $zipName ]
	# запустить zip
	puts "Create archive $zipName"
	set owd [ pwd ]
	cd $dataDir
	if { [ file exists $zipName ] } {
		file delete $zipName
	}
	if { [ catch { exec {*}$cmdline << [ join $fileList "\n" ] } msg ] } {
		puts $msg
	}
	cd $owd
}

##################
if { $argv == {} } {
	package require Tk
	wm withdraw .
	update
#	catch {
		tk appname "mq45dist"
		package require tkcon
		set tkcon::PRIV(root) .console
		tkcon::Init
		tkcon titla "mq45dist"
		tkcon show
#	}
	
	set fileName [ tk_getOpenFile -title "Choose any file from your project" ]
	if { $fileName == {} } {
		exit 0
	}
	mq45dist $fileName
	tkwait window .console
} else {
	puts "from cmdline"
	mq45dist [ join $argv " " ]
}
exit 0
